@@ -68,6 +68,7 @@ typedef std::auto_ptr<Geometry> geom_ptr;
 static std::vector<std::string> wkts;
 static std::vector<double> areas;
 
+static int excludepoly = 0;
 
 char *get_wkt_simple(osmNode *nodes, int count, int polygon) {
     GeometryFactory gf;
@@ -86,6 +87,13 @@ char *get_wkt_simple(osmNode *nodes, int count, int polygon) {
         if (polygon && (coords->getSize() >= 4) && (coords->getAt(coords->getSize() - 1).equals2D(coords->getAt(0)))) {
             std::auto_ptr<LinearRing> shell(gf.createLinearRing(coords.release()));
             geom = geom_ptr(gf.createPolygon(shell.release(), new std::vector<Geometry *>));
+            if (!geom->isValid()) {
+                if (excludepoly) {
+                    return NULL;
+                } else {   
+                    geom = geom_ptr(geom->buffer(0));
+                }
+            }
             geom->normalize(); // Fix direction of ring
         } else {
             if (coords->getSize() < 2)
@@ -131,6 +139,13 @@ size_t get_wkt_split(osmNode *nodes, int count, int polygon, double split_at) {
         if (polygon && (coords->getSize() >= 4) && (coords->getAt(coords->getSize() - 1).equals2D(coords->getAt(0)))) {
             std::auto_ptr<LinearRing> shell(gf.createLinearRing(coords.release()));
             geom = geom_ptr(gf.createPolygon(shell.release(), new std::vector<Geometry *>));
+            if (!geom->isValid()) {
+                if (excludepoly) {
+                    return 0;
+                } else {
+                    geom = geom_ptr(geom->buffer(0));
+                }
+            }
             geom->normalize(); // Fix direction of ring
             area = geom->getArea();
             std::string wkt = wktw.write(geom.get());
@@ -474,21 +489,36 @@ size_t build_geometry(osmid_t osm_id, struct osmNode **xnodes, int *xcount, int
             // Make a multipolygon if required
             if ((toplevelpolygons > 1) && enable_multi)
             {
-                std::auto_ptr<MultiPolygon> multipoly(gf.createMultiPolygon(polygons.release()));
-                std::string text = writer.write(multipoly.get());
-                wkts.push_back(text);
-                areas.push_back(multipoly->getArea());
-                wkt_size++;
+                geom_ptr multipoly(gf.createMultiPolygon(polygons.release()));
+                if (!multipoly->isValid() && (excludepoly == 0)) {
+                    multipoly = geom_ptr(multipoly->buffer(0));
+                }
+                multipoly->normalize();
+
+                if ((excludepoly == 0) || (multipoly->isValid()))
+                {
+                    std::string text = writer.write(multipoly.get());
+                    wkts.push_back(text);
+                    areas.push_back(multipoly->getArea());
+                    wkt_size++;
+                }
             }
             else
             {
                 for(unsigned i=0; i<toplevelpolygons; i++) 
                 {
-                    Polygon* poly = dynamic_cast<Polygon*>(polygons->at(i));;
-                    std::string text = writer.write(poly);
-                    wkts.push_back(text);
-                    areas.push_back(poly->getArea());
-                    wkt_size++;
+                    Geometry* poly = dynamic_cast<Geometry*>(polygons->at(i));
+                    if (!poly->isValid() && (excludepoly == 0)) {
+                        poly = dynamic_cast<Geometry*>(poly->buffer(0));
+                        poly->normalize();
+                    }
+                    if ((excludepoly == 0) || (poly->isValid()))
+                    {
+                        std::string text = writer.write(poly);
+                        wkts.push_back(text);
+                        areas.push_back(poly->getArea());
+                        wkt_size++;
+                    }
                     delete(poly);
                 }
             }
@@ -513,3 +543,8 @@ size_t build_geometry(osmid_t osm_id, struct osmNode **xnodes, int *xcount, int
 
     return wkt_size;
 }
+
+void exclude_broken_polygon ()
+{
+    excludepoly = 1;
+}
