@@ -11,6 +11,11 @@
 struct keyval {
     char *key;
     char *value;
+    /* if a hstore column is requested we need a flag to store if a key
+       has its own column because it should not be added to the hstore 
+       in this case
+    */
+    int has_column;
     struct keyval *next;
     struct keyval *prev;
 };
@@ -20,6 +25,8 @@ void freeItem(struct keyval *p);
 unsigned int countList(struct keyval *head);
 int listHasData(struct keyval *head);
 char *getItem(struct keyval *head, const char *name);
+struct keyval *getTag(struct keyval *head, const char *name);
+void removeTag(struct keyval *tag);
 struct keyval *firstItem(struct keyval *head);
 struct keyval *nextItem(struct keyval *head, struct keyval *item);
 struct keyval *popItem(struct keyval *head);
