@@ -18,13 +18,6 @@
 
 
 
-/* Store +-20,000km Mercator co-ordinates as fixed point 32bit number with maximum precision */
-/* Scale is chosen such that 40,000 * SCALE < 2^32          */
-#define FIXED_POINT
-
-static int scale = 100;
-#define DOUBLE_TO_FIX(x) ((int)((x) * scale))
-#define FIX_TO_DOUBLE(x) (((double)x) / scale)
 
 /* Here we use a similar storage structure as middle-ram, except we allow
  * the array to be lossy so we can cap the total memory usage. Hence it is a
@@ -63,26 +56,7 @@ static int scale = 100;
  *  Reuse old block: O(log maxBlocks)
  */
 
-struct ramNode {
-#ifdef FIXED_POINT
-    int lon;
-    int lat;
-#else
-    double lon;
-    double lat;
-#endif
-};
 
-struct ramNodeID {
-    osmid_t id;
-    struct ramNode coord;
-};
-
-struct ramNodeBlock {
-    struct ramNode    *nodes;
-    int block_offset;
-    int used;
-};
 
 static int allocStrategy = ALLOC_DENSE;
 
@@ -90,7 +64,7 @@ static int allocStrategy = ALLOC_DENSE;
 #define PER_BLOCK  (((osmid_t)1) << BLOCK_SHIFT)
 #define NUM_BLOCKS (((osmid_t)1) << (36 - BLOCK_SHIFT))
 
-#define SAFETY_MARGIN 1024*1024
+#define SAFETY_MARGIN 1024*PER_BLOCK*sizeof(struct ramNode)
 
 static struct ramNodeBlock *blocks;
 static int usedBlocks;
@@ -114,23 +88,23 @@ static int warn_node_order;
 
 static int ram_cache_nodes_get_sparse(struct osmNode *out, osmid_t id);
 
-static inline int id2block(osmid_t id)
+static int id2block(osmid_t id)
 {
-    // + NUM_BLOCKS/2 allows for negative IDs
+    /* + NUM_BLOCKS/2 allows for negative IDs */
     return (id >> BLOCK_SHIFT) + NUM_BLOCKS/2;
 }
 
-static inline int id2offset(osmid_t id)
+static int id2offset(osmid_t id)
 {
     return id & (PER_BLOCK-1);
 }
 
-static inline osmid_t block2id(int block, int offset)
+static osmid_t block2id(int block, int offset)
 {
     return (((osmid_t) block - NUM_BLOCKS/2) << BLOCK_SHIFT) + (osmid_t) offset;
 }
 
-#define Swap(a,b) { typeof(a) __tmp = a; a = b; b = __tmp; }
+#define Swap(a,b) { struct ramNodeBlock * __tmp = a; a = b; b = __tmp; }
 
 static void percolate_up( int pos )
 {
@@ -140,7 +114,7 @@ static void percolate_up( int pos )
       int parent = (i-1)>>1;
       if( queue[i]->used < queue[parent]->used )
       {
-        Swap( queue[i], queue[parent] );
+        Swap( queue[i], queue[parent] )
         i = parent;
       }
       else
@@ -260,7 +234,7 @@ static int ram_cache_nodes_set_dense(osmid_t id, double lat, double lon, struct
              * current head of the tree down to the right level to restore the
              * priority queue invariant. Upto log(maxBlocks) iterations */
             
-            int i=0;
+            i=0;
             while( 2*i+1 < usedBlocks - 1 ) {
                 if( queue[2*i+1]->used <= queue[2*i+2]->used ) {
                     if( queue[i]->used > queue[2*i+1]->used ) {
@@ -387,8 +361,12 @@ void init_node_ram_cache( int strategy, int cacheSizeMB, int fixpointscale ) {
     
     if ((allocStrategy & ALLOC_DENSE) > 0 ) {
         fprintf(stderr, "Allocating memory for dense node cache\n");
-        blocks = malloc(sizeof(struct ramNodeBlock)*NUM_BLOCKS);
-        queue = malloc( maxBlocks * sizeof(struct ramNodeBlock) );
+        blocks = calloc(NUM_BLOCKS,sizeof(struct ramNodeBlock));
+        if (!blocks) {
+            fprintf(stderr, "Out of memory for node cache dense index, try using \"--cache-strategy sparse\" instead \n");
+            exit_nicely();
+        }
+        queue = calloc( maxBlocks,sizeof(struct ramNodeBlock *) );
         /* Use this method of allocation if virtual memory is limited,
          * or if OS allocs physical memory right away, rather than page by page
          * once it is needed.
@@ -401,7 +379,7 @@ void init_node_ram_cache( int strategy, int cacheSizeMB, int fixpointscale ) {
             }
         } else {
             fprintf(stderr, "Allocating dense node cache in one big chunk\n");
-            blockCache = malloc(maxBlocks * PER_BLOCK * sizeof(struct ramNode) + SAFETY_MARGIN);
+            blockCache = calloc(maxBlocks + 1024,PER_BLOCK * sizeof(struct ramNode));
             if (!queue || !blockCache) {
                 fprintf(stderr, "Out of memory for dense node cache, reduce --cache size\n");
                 exit_nicely();
@@ -419,7 +397,7 @@ void init_node_ram_cache( int strategy, int cacheSizeMB, int fixpointscale ) {
     if ((allocStrategy & ALLOC_SPARSE) > 0 ) {
         fprintf(stderr, "Allocating memory for sparse node cache\n");
         if (!blockCache) {
-            sparseBlock = malloc(maxSparseTuples * sizeof(struct ramNodeID));
+            sparseBlock = calloc(maxSparseTuples,sizeof(struct ramNodeID));
         } else {
             fprintf(stderr, "Sharing dense sparse\n");
             sparseBlock = blockCache;
