@@ -25,7 +25,6 @@
 
 #include "config.h"
 
-#define _GNU_SOURCE
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -55,6 +54,7 @@
 #include "sprompt.h"
 #include "parse-xml2.h"
 #include "parse-primitive.h"
+#include "parse-o5m.h"
 
 #ifdef BUILD_READER_PBF
 #  include "parse-pbf.h"
@@ -65,7 +65,7 @@
 
 int verbose;
 
-// Data structure carrying all parsing related variables
+/* Data structure carrying all parsing related variables */
 static struct osmdata_t osmdata = { 
   .filetype = FILETYPE_NONE,
   .action   = ACTION_NONE,
@@ -93,7 +93,7 @@ static int parse_bbox(struct osmdata_t *osmdata)
         fprintf(stderr, "Bounding box failed due to maxlat <= minlat\n");
         return 1;
     }
-    printf("Applying Bounding box: %f,%f to %f,%f\n", osmdata->minlon, osmdata->minlat, osmdata->maxlon, osmdata->maxlat);
+    fprintf(stderr, "Applying Bounding box: %f,%f to %f,%f\n", osmdata->minlon, osmdata->minlat, osmdata->maxlon, osmdata->maxlat);
     return 0;
 }
 
@@ -187,18 +187,24 @@ static void long_usage(char *arg0)
     printf("              \t\tThis includes the username, userid, timestamp and version.\n"); 
     printf("              \t\tNote: this option also requires additional entries in your style file.\n"); 
     printf("   -k|--hstore\t\tAdd tags without column to an additional hstore (key/value) column to postgresql tables\n");
+    printf("      --hstore-match-only\tOnly keep objects that have a value in one of the columns\n");
+    printf("      -                  \t(normal action with --hstore is to keep all objects)\n");
     printf("   -j|--hstore-all\tAdd all tags to an additional hstore (key/value) column in postgresql tables\n");
     printf("   -z|--hstore-column\tAdd an additional hstore (key/value) column containing all tags\n");
     printf("                     \tthat start with the specified string, eg --hstore-column \"name:\" will\n");
     printf("                     \tproduce an extra hstore column that contains all name:xx tags\n");
+    printf("      --hstore-add-index\tAdd index to hstore column.\n");
     printf("   -G|--multi-geometry\tGenerate multi-geometry features in postgresql tables.\n");
     printf("   -K|--keep-coastlines\tKeep coastline data rather than filtering it out.\n");
     printf("              \t\tBy default natural=coastline tagged data will be discarded based on the\n");
     printf("              \t\tassumption that post-processed Coastline Checker shapefiles will be used.\n");
+    printf("      --exclude-invalid-polygon\n");
+#ifdef HAVE_FORK
     printf("      --number-processes\t\tSpecifies the number of parallel processes used for certain operations\n");
     printf("             \t\tDefault is 1\n");
+#endif
     printf("   -I|--disable-parallel-indexing\tDisable indexing all tables concurrently.\n");
-    printf("      --unlogged-tables\tUse unlogged tables (lost on crash but faster). Requires PostgreSQL 9.1.\n");
+    printf("      --unlogged\tUse unlogged tables (lost on crash but faster). Requires PostgreSQL 9.1.\n");
     printf("      --cache-strategy\tSpecifies the method used to cache nodes in ram.\n");
     printf("                      \t\tAvailable options are:\n");
     printf("                      \t\tdense: caching strategy optimised for full planet import\n");
@@ -206,13 +212,16 @@ static void long_usage(char *arg0)
     printf("                      \t\tsparse: caching strategy optimised for small extracts\n");
     printf("                      \t\toptimized: automatically combines dense and sparse strategies for optimal storage efficiency.\n");
     printf("                      \t\t           optimized may use twice as much virtual memory, but no more physical memory\n");
+
 #ifdef __amd64__
     printf("                      \t\t   The default is \"optimized\"\n");
 #else
     /* use "chunked" as a default in 32 bit compilations, as it is less wasteful of virtual memory than "optimized"*/
-    printf("                      \t\t   The default is \"chunked\"\n");
+    printf("                      \t\t   The default is \"sparse\"\n");
 #endif
-
+    printf("      --flat-nodes\tSpecifies the flat file to use to persistently store node information in slim mode instead of in pgsql\n");
+    printf("                  \t\tThis file is a single > 16Gb large file. This method is only recomended for full planet imports\n");
+    printf("                   \t\tas it doesn't work well with small extracts. The default is disabled\n");
     printf("   -h|--help\t\tHelp information.\n");
     printf("   -v|--verbose\t\tVerbose output.\n");
     printf("\n");
@@ -297,17 +306,21 @@ void realloc_members(struct osmdata_t *osmdata)
 
 void resetMembers(struct osmdata_t *osmdata)
 {
-  for(unsigned i = 0; i < osmdata->member_count; i++ )
+  unsigned i;
+  for(i = 0; i < osmdata->member_count; i++ )
     free( osmdata->members[i].role );
 }
 
 void printStatus(struct osmdata_t *osmdata)
 {
     time_t now;
+    time_t end_nodes;
+    time_t end_way;
+    time_t end_rel;
     time(&now);
-    time_t end_nodes = osmdata->start_way > 0 ? osmdata->start_way : now;
-    time_t end_way = osmdata->start_rel > 0 ? osmdata->start_rel : now;
-    time_t end_rel =  now;
+    end_nodes = osmdata->start_way > 0 ? osmdata->start_way : now;
+    end_way = osmdata->start_rel > 0 ? osmdata->start_rel : now;
+    end_rel =  now;
     fprintf(stderr, "\rProcessing: Node(%" PRIdOSMID "k %.1fk/s) Way(%" PRIdOSMID "k %.2fk/s) Relation(%" PRIdOSMID " %.2f/s)",
             osmdata->count_node/1000,
             (double)osmdata->count_node/1000.0/((int)(end_nodes - osmdata->start_node) > 0 ? (double)(end_nodes - osmdata->start_node) : 1.0),
@@ -343,26 +356,36 @@ int main(int argc, char *argv[])
     int expire_tiles_zoom = -1;
     int expire_tiles_zoom_min = -1;
     int enable_hstore = HSTORE_NONE;
+    int enable_hstore_index = 0;
+    int hstore_match_only = 0;
     int enable_multi = 0;
     int parallel_indexing = 1;
+    int flat_node_cache_enabled = 0;
 #ifdef __amd64__
     int alloc_chunkwise = ALLOC_SPARSE | ALLOC_DENSE;
 #else
-    int alloc_chunkwise = ALLOC_DENSE_CHUNK | ALLOC_DENSE;
+    int alloc_chunkwise = ALLOC_SPARSE;
 #endif
     int num_procs = 1;
     int droptemp = 0;
     int unlogged = 0;
+    int excludepoly = 0;
+    time_t start, end;
+    time_t overall_start, overall_end;
+    time_t now;
+    time_t end_nodes;
+    time_t end_way;
+    time_t end_rel;
     const char *expire_tiles_filename = "dirty_tiles";
     const char *db = "gis";
     const char *username=NULL;
     const char *host=NULL;
     const char *password=NULL;
     const char *port = "5432";
-    const char *tblsmain_index = NULL; // no default TABLESPACE for index on main tables
-    const char *tblsmain_data = NULL;  // no default TABLESPACE for main tables
-    const char *tblsslim_index = NULL; // no default TABLESPACE for index on slim mode tables
-    const char *tblsslim_data = NULL;  // no default TABLESPACE for slim mode tables
+    const char *tblsmain_index = NULL; /* no default TABLESPACE for index on main tables */
+    const char *tblsmain_data = NULL;  /* no default TABLESPACE for main tables */
+    const char *tblsslim_index = NULL; /* no default TABLESPACE for index on slim mode tables */
+    const char *tblsslim_data = NULL;  /* no default TABLESPACE for slim mode tables */
     const char *conninfo = NULL;
     const char *prefix = "planet_osm";
     const char *style = OSM2PGSQL_DATADIR "/default.style";
@@ -370,6 +393,7 @@ int main(int argc, char *argv[])
     const char *output_backend = "pgsql";
     const char *input_reader = "auto";
     const char **hstore_columns = NULL;
+    const char *flat_nodes_file = NULL;
     int n_hstore_columns = 0;
     int keep_coastlines=0;
     int cache = 800;
@@ -378,7 +402,7 @@ int main(int argc, char *argv[])
     
     int (*streamFile)(char *, int, struct osmdata_t *);
 
-    printf("osm2pgsql SVN version %s (%lubit id space)\n\n", VERSION, 8 * sizeof(osmid_t));
+    fprintf(stderr, "osm2pgsql SVN version %s (%lubit id space)\n\n", VERSION, 8 * sizeof(osmid_t));
 
     while (1) {
         int c, option_index = 0;
@@ -414,6 +438,8 @@ int main(int argc, char *argv[])
             {"hstore", 0, 0, 'k'},
             {"hstore-all", 0, 0, 'j'},
             {"hstore-column", 1, 0, 'z'},
+            {"hstore-match-only", 0, 0, 208},
+            {"hstore-add-index",0,0,211},
             {"multi-geometry", 0, 0, 'G'},
             {"keep-coastlines", 0, 0, 'K'},
             {"input-reader", 1, 0, 'r'},
@@ -423,6 +449,8 @@ int main(int argc, char *argv[])
             {"number-processes", 1, 0, 205},
             {"drop", 0, 0, 206},
             {"unlogged", 0, 0, 207},
+            {"flat-nodes",1,0,209},
+            {"exclude-invalid-polygon",0,0,210},
             {0, 0, 0, 0}
         };
 
@@ -465,10 +493,11 @@ int main(int argc, char *argv[])
             case 'O': output_backend = optarg; break;
             case 'x': osmdata.extra_attributes=1; break;
             case 'k': enable_hstore=HSTORE_NORM; break;
-	    case 'j': enable_hstore=HSTORE_ALL; break;
+            case 208: hstore_match_only = 1; break;
+            case 'j': enable_hstore=HSTORE_ALL; break;
             case 'z': 
                 n_hstore_columns++;
-                hstore_columns = (const char**)realloc(hstore_columns, sizeof(&n_hstore_columns) * n_hstore_columns);
+                hstore_columns = (const char**)realloc(hstore_columns, sizeof(char *) * n_hstore_columns);
                 hstore_columns[n_hstore_columns-1] = optarg;
                 break;
             case 'G': enable_multi=1; break;
@@ -485,9 +514,21 @@ int main(int argc, char *argv[])
                 if (strcmp(optarg,"sparse") == 0) alloc_chunkwise = ALLOC_SPARSE;
                 if (strcmp(optarg,"optimized") == 0) alloc_chunkwise = ALLOC_DENSE | ALLOC_SPARSE;
                 break;
-            case 205: num_procs = atoi(optarg); break;
+            case 205:
+#ifdef HAVE_FORK                
+                num_procs = atoi(optarg);
+#else
+                fprintf(stderr, "WARNING: osm2pgsql was compiled without fork, only using one process!\n");
+#endif
+                break;
             case 206: droptemp = 1; break;
             case 207: unlogged = 1; break;
+            case 209:
+            	flat_node_cache_enabled = 1;
+            	flat_nodes_file = optarg;
+            	break;
+            case 210: excludepoly = 1; exclude_broken_polygon(); break;
+            case 211: enable_hstore_index = 1; break;
             case 'V': exit(EXIT_SUCCESS);
             case '?':
             default:
@@ -501,7 +542,7 @@ int main(int argc, char *argv[])
         exit(EXIT_SUCCESS);
     }
 
-    if (argc == optind) {  // No non-switch arguments
+    if (argc == optind) {  /* No non-switch arguments */
         short_usage(argv[0]);
         exit(EXIT_FAILURE);
     }
@@ -521,6 +562,17 @@ int main(int argc, char *argv[])
         unlogged = 0;
     }
 
+    if (enable_hstore == HSTORE_NONE && !n_hstore_columns && hstore_match_only)
+    {
+        fprintf(stderr, "Warning: --hstore-match-only only makes sense with --hstore, --hstore-all, or --hstore-column; ignored.\n");
+        hstore_match_only = 0;
+    }
+
+    if (enable_hstore_index && enable_hstore  == HSTORE_NONE && !n_hstore_columns) {
+        fprintf(stderr, "Warning: --hstore-add-index only makes sense with hstore enabled.\n");
+        enable_hstore_index = 0;
+    }
+
     if (cache < 0) cache = 0;
 
     if (num_procs < 1) num_procs = 1;
@@ -529,7 +581,9 @@ int main(int argc, char *argv[])
         password = simple_prompt("Password:", 100, 0);
     else {
         password = getenv("PGPASS");
-    }	
+    }
+
+    
 
     conninfo = build_conninfo(db, username, password, host, port);
     sql_conn = PQconnectdb(conninfo);
@@ -579,6 +633,8 @@ int main(int argc, char *argv[])
     options.expire_tiles_filename = expire_tiles_filename;
     options.enable_multi = enable_multi;
     options.enable_hstore = enable_hstore;
+    options.enable_hstore_index = enable_hstore_index;
+    options.hstore_match_only = hstore_match_only;
     options.hstore_columns = hstore_columns;
     options.n_hstore_columns = n_hstore_columns;
     options.keep_coastlines = keep_coastlines;
@@ -587,6 +643,9 @@ int main(int argc, char *argv[])
     options.num_procs = num_procs;
     options.droptemp = droptemp;
     options.unlogged = unlogged;
+    options.flat_node_cache_enabled = flat_node_cache_enabled;
+    options.flat_node_file = flat_nodes_file;
+    options.excludepoly = excludepoly;
 
     if (strcmp("pgsql", output_backend) == 0) {
       osmdata.out = &out_pgsql;
@@ -609,8 +668,10 @@ int main(int argc, char *argv[])
       } else if (strcmp("pbf", input_reader) == 0) {
         streamFile = &streamFilePbf;
 #endif
+      } else if (strcmp("o5m", input_reader) == 0) {
+          streamFile = &streamFileO5m;
       } else {
-        fprintf(stderr, "Input parser `%s' not recognised. Should be one of [libxml2, primitive"
+        fprintf(stderr, "Input parser `%s' not recognised. Should be one of [libxml2, primitive, o5m"
 #ifdef BUILD_READER_PBF
 	      ", pbf"
 #endif
@@ -619,9 +680,7 @@ int main(int argc, char *argv[])
       }
     }
 
-    time_t overall_start, overall_end;
     time(&overall_start);
-
     osmdata.out->start(&options);
 
     realloc_nodes(&osmdata);
@@ -638,18 +697,20 @@ int main(int argc, char *argv[])
         /* if input_reader is not forced by -r switch try to auto-detect it
            by file extension */
         if (strcmp("auto", input_reader) == 0) {
-#ifdef BUILD_READER_PBF
+
           if (strcasecmp(".pbf",argv[optind]+strlen(argv[optind])-4) == 0) {
+#ifdef BUILD_READER_PBF
             streamFile = &streamFilePbf;
+#else
+	    fprintf(stderr, "ERROR: PBF support has not been compiled into this version of osm2pgsql, please either compile it with pbf support or use one of the other input formats\n");
+	    exit(EXIT_FAILURE);
+#endif
+          } else if (strcasecmp(".o5m",argv[optind]+strlen(argv[optind])-4) == 0) {
+              streamFile = &streamFileO5m;
           } else {
             streamFile = &streamFileXML2;
           }
-#else
-          streamFile = &streamFileXML2;
-#endif
         }
-        time_t start, end;
-
         fprintf(stderr, "\nReading in file: %s\n", argv[optind]);
         time(&start);
         if (streamFile(argv[optind], sanitize, &osmdata) != 0)
@@ -663,11 +724,10 @@ int main(int argc, char *argv[])
     xmlMemoryDump();
     
     if (osmdata.count_node || osmdata.count_way || osmdata.count_rel) {
-        time_t now;
         time(&now);
-        time_t end_nodes = osmdata.start_way > 0 ? osmdata.start_way : now;
-        time_t end_way = osmdata.start_rel > 0 ? osmdata.start_rel : now;
-        time_t end_rel =  now;
+        end_nodes = osmdata.start_way > 0 ? osmdata.start_way : now;
+        end_way = osmdata.start_rel > 0 ? osmdata.start_rel : now;
+        end_rel =  now;
         fprintf(stderr, "\n");
         fprintf(stderr, "Node stats: total(%" PRIdOSMID "), max(%" PRIdOSMID ") in %is\n", osmdata.count_node, osmdata.max_node,
                 osmdata.count_node > 0 ? (int)(end_nodes - osmdata.start_node) : 0);
@@ -681,7 +741,7 @@ int main(int argc, char *argv[])
     free(osmdata.nds);
     free(osmdata.members);
     
-    // free the column pointer buffer
+    /* free the column pointer buffer */
     free(hstore_columns);
 
     project_exit();
