@@ -28,30 +28,10 @@
    "  score FLOAT"                               \
    ")"
 
-#define CREATE_PLACE_TABLE                      \
-   "CREATE TABLE place ("                       \
-   "  place_id BIGINT,"                         \
-   "  osm_type CHAR(1) NOT NULL,"               \
-   "  osm_id BIGINT NOT NULL,"                  \
-   "  class TEXT NOT NULL,"                     \
-   "  type TEXT NOT NULL,"                      \
-   "  name keyvalue[],"                         \
-   "  admin_level INTEGER,"                     \
-   "  housenumber TEXT,"                        \
-   "  street TEXT,"                             \
-   "  isin TEXT,"                               \
-   "  postcode TEXT,"                           \
-   "  country_code VARCHAR(2),"                 \
-   "  street_place_id BIGINT,"                  \
-   "  rank_address INTEGER,"                    \
-   "  rank_search INTEGER,"                     \
-   "  indexed BOOLEAN"                          \
-   ") %s %s"
-
-#define V2_CREATE_PLACE_TABLE                   \
+#define CREATE_PLACE_TABLE                   \
    "CREATE TABLE place ("                       \
    "  osm_type CHAR(1) NOT NULL,"               \
-   "  osm_id BIGINT NOT NULL,"                  \
+   "  osm_id " POSTGRES_OSMID_TYPE " NOT NULL," \
    "  class TEXT NOT NULL,"                     \
    "  type TEXT NOT NULL,"                      \
    "  name HSTORE,"                             \
@@ -73,35 +53,6 @@
 #define TAGINFO_WAY  0x2u
 #define TAGINFO_AREA 0x4u
 
-static const struct taginfo {
-   const char   *name;
-   const char   *value;
-   unsigned int flags;
-} taginfo[] = {
-   { "aeroway",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "amenity",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "boundary",  NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "bridge",    NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "building",  NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "craft",     NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "emergency", NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "highway",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "historic",  NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "landuse",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "leisure",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "military",  NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "natural",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "office",    NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "place",     NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "railway",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "shop",      NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "tourism",   NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "tunnel",    NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { "waterway",  NULL,             TAGINFO_NODE|TAGINFO_WAY|TAGINFO_AREA },
-   { NULL,        NULL,             0                                     }
-};
-
-//static int gazetteer_delete_relation(osmid_t osm_id);
 
 static const struct output_options *Options = NULL;
 static PGconn *Connection = NULL;
@@ -109,6 +60,8 @@ static int CopyActive = 0;
 static char Buffer[BUFFER_SIZE];
 static unsigned int BufferLen = 0;
 
+static PGconn *ConnectionDelete = NULL;
+
 static PGconn *ConnectionError = NULL;
 static int CopyErrorActive = 0;
 static char BufferError[BUFFER_SIZE];
@@ -282,12 +235,16 @@ static void stop_error_copy(void)
 }
 
 static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *names, struct keyval *places, struct keyval *extratags, 
-   int* admin_level, char ** housenumber, char ** street, char ** isin, char ** postcode, char ** countrycode)
+   int* admin_level, struct keyval ** housenumber, struct keyval ** street, char ** isin, struct keyval ** postcode, struct keyval ** countrycode)
 {
    int placehouse = 0;
    int placebuilding = 0;
-   char landuse[256];
+   int placeadmin = 0;
+   struct keyval *landuse;
+   struct keyval *place;
    struct keyval *item;
+   struct keyval *conscriptionnumber;
+   struct keyval *streetnumber;
 
    *admin_level = ADMINLEVEL_NONE;
    *housenumber = 0;
@@ -296,7 +253,10 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
    int isinsize = 0;
    *postcode = 0;
    *countrycode = 0;
-   *landuse = 0;
+   landuse = 0;
+   place = 0;
+   conscriptionnumber = 0;
+   streetnumber = 0;
 
    /* Initialise the result lists */
    initList(names);
@@ -306,7 +266,6 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
    /* Loop over the tags */
    while ((item = popItem(tags)) != NULL)
    {
-//      fprintf(stderr, "%s\n", item->key);
 
       /* If this is a name tag, add it to the name list */
       if (strcmp(item->key, "ref") == 0 ||
@@ -350,11 +309,19 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
           (strncmp(item->key, "place_name:", 11) == 0) ||
           strcmp(item->key, "short_name") == 0 ||
           (strncmp(item->key, "short_name:", 11) == 0) ||
-          strcmp(item->key, "operator") == 0) //operator is a bit of an oddity
+          strcmp(item->key, "operator") == 0) /* operator is a bit of an oddity */
       {
-         pushItem(names, item);
+         if (strcmp(item->key, "name:prefix") == 0)
+         {
+            pushItem(extratags, item);
+         }
+         else
+         {
+            pushItem(names, item);
+         }
       }
-      else if (strcmp(item->key, "aeroway") == 0 ||
+      else if (strcmp(item->key, "aerialway") == 0 ||
+               strcmp(item->key, "aeroway") == 0 ||
                strcmp(item->key, "amenity") == 0 ||
                strcmp(item->key, "boundary") == 0 ||
                strcmp(item->key, "bridge") == 0 ||
@@ -366,14 +333,28 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
                strcmp(item->key, "military") == 0 ||
                strcmp(item->key, "natural") == 0 ||
                strcmp(item->key, "office") == 0 ||
-               strcmp(item->key, "place") == 0 ||
                strcmp(item->key, "railway") == 0 ||
                strcmp(item->key, "shop") == 0 ||
                strcmp(item->key, "tourism") == 0 ||
                strcmp(item->key, "tunnel") == 0 ||
                strcmp(item->key, "waterway") == 0 )
       {
-         pushItem(places, item);
+         if (strcmp(item->value, "no"))
+         {
+            pushItem(places, item);
+            if (strcmp(item->key, "boundary") == 0 && strcmp(item->value, "administrative") == 0)
+            {
+               placeadmin = 1;
+            }
+         }
+         else
+         {
+            freeItem(item);
+         }
+      }
+      else if (strcmp(item->key, "place") == 0) 
+      {
+         place = item;
       }
       else if (strcmp(item->key, "addr:housename") == 0)
       {
@@ -382,8 +363,10 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
       }
       else if (strcmp(item->key, "landuse") == 0)
       {
-         strcpy(landuse, item->value);
-         freeItem(item);
+         if (strcmp(item->value, "cemetery") == 0)
+            pushItem(places, item);
+         else
+            landuse = item;
       }
       else if (strcmp(item->key, "postal_code") == 0 ||
           strcmp(item->key, "post_code") == 0 ||
@@ -393,18 +376,19 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
           strcmp(item->key, "tiger:zip_right") == 0)
       {
          if (*postcode)
-	    freeItem(item);
+	        freeItem(item);
          else
-            *postcode = item->value;
+            *postcode = item;
       }
       else if (strcmp(item->key, "addr:street") == 0)
       {
-         *street = item->value;
+         *street = item;
       }
       else if ((strcmp(item->key, "country_code_iso3166_1_alpha_2") == 0 || 
                 strcmp(item->key, "country_code_iso3166_1") == 0 || 
                 strcmp(item->key, "country_code_iso3166") == 0 || 
                 strcmp(item->key, "country_code") == 0 || 
+                strcmp(item->key, "iso3166-1:alpha2") == 0 || 
                 strcmp(item->key, "iso3166-1") == 0 || 
                 strcmp(item->key, "ISO3166-1") == 0 || 
                 strcmp(item->key, "iso3166") == 0 || 
@@ -413,19 +397,45 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
                 strcmp(item->key, "addr:country_code") == 0) 
                 && strlen(item->value) == 2)
       {
-         *countrycode = item->value;
+         *countrycode = item;
       }
       else if (strcmp(item->key, "addr:housenumber") == 0)
       {
-         // house number can be far more complex than just a single house number - leave for postgresql to deal with
-         *housenumber = item->value; 
-         placehouse = 1;
+          /* house number can be far more complex than just a single house number - leave for postgresql to deal with */
+         if (*housenumber)
+             freeItem(item);
+         else {
+             *housenumber = item;
+             placehouse = 1;
+         }
+      }
+      else if (strcmp(item->key, "addr:conscriptionnumber") == 0)
+      {
+         if (conscriptionnumber)
+             freeItem(item);
+         else {
+             conscriptionnumber = item;
+             placehouse = 1;
+         }
+      }
+      else if (strcmp(item->key, "addr:streetnumber") == 0)
+      {
+         if (streetnumber)
+             freeItem(item);
+         else {
+             streetnumber = item;
+             placehouse = 1;
+         }
       }
       else if (strcmp(item->key, "addr:interpolation") == 0)
       {
-         // house number can be far more complex than just a single house number - leave for postgresql to deal with
-         *housenumber = item->value; 
-         addItem(places, "place", "houses", 1);
+          /* house number can be far more complex than just a single house number - leave for postgresql to deal with */
+          if (*housenumber) {
+              freeItem(item);
+          } else {
+             *housenumber = item; 
+             addItem(places, "place", "houses", 1);
+          }
       }
       else if (strcmp(item->key, "is_in") == 0 ||
           (strncmp(item->key, "is_in:", 5) == 0) ||
@@ -440,17 +450,19 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
          *(*isin+isinsize) = ',';
          strcpy(*isin+1+isinsize, item->value);
          isinsize += 1 + strlen(item->value);
+         freeItem(item);
       }
       else if (strcmp(item->key, "admin_level") == 0)
       {
          *admin_level = atoi(item->value);
+         freeItem(item);
       }
       else if (strcmp(item->key, "tracktype") == 0 ||
                strcmp(item->key, "traffic_calming") == 0 ||
                strcmp(item->key, "service") == 0 ||
                strcmp(item->key, "cuisine") == 0 ||
                strcmp(item->key, "capital") == 0 ||
-               strcmp(item->key, "dispending") == 0 ||
+               strcmp(item->key, "dispensing") == 0 ||
                strcmp(item->key, "religion") == 0 ||
                strcmp(item->key, "denomination") == 0 ||
                strcmp(item->key, "sport") == 0 ||
@@ -495,7 +507,7 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
                strcmp(item->key, "hour_off") == 0 ||
                strcmp(item->key, "maxweight") == 0 ||
                strcmp(item->key, "maxheight") == 0 ||
-               strcmp(item->key, "speed") == 0 ||
+               strcmp(item->key, "maxspeed") == 0 ||
                strcmp(item->key, "disused") == 0 ||
                strcmp(item->key, "toll") == 0 ||
                strcmp(item->key, "charge") == 0 ||
@@ -508,20 +520,27 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
                strcmp(item->key, "url") == 0 ||
                strcmp(item->key, "website") == 0 ||
                strcmp(item->key, "phone") == 0 ||
-               strcmp(item->key, "phone") == 0 ||
                strcmp(item->key, "tel") == 0 ||
                strcmp(item->key, "real_ale") == 0 ||
                strcmp(item->key, "smoking") == 0 ||
                strcmp(item->key, "food") == 0 ||
-               strcmp(item->key, "camra") == 0 ||
+               strcmp(item->key, "camera") == 0 ||
                strcmp(item->key, "brewery") == 0 ||
-               strcmp(item->key, "wikipedia") == 0)
+               strcmp(item->key, "locality") == 0 ||
+               strcmp(item->key, "wikipedia") == 0 ||
+               (strncmp(item->key, "wikipedia:", 10) == 0)
+               )
       {
           pushItem(extratags, item);
       }
       else if (strcmp(item->key, "building") == 0)
       {
           placebuilding = 1;
+          freeItem(item);
+      }
+      else if (strcmp(item->key, "mountain_pass") == 0)
+      {
+          pushItem(places, item);
       }
       else
       {
@@ -529,20 +548,78 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
       }
    }
 
+   /* Handle Czech/Slovak addresses:
+        - if we have just a conscription number or a street number,
+          just use the one we have as a house number
+        - if we have both of them, concatenate them so users may search
+          by any of them
+    */
+   if (conscriptionnumber || streetnumber)
+   {
+      if (*housenumber)
+      {
+         freeItem(*housenumber);
+      }
+      if (!conscriptionnumber)
+      {
+         addItem(tags, "addr:housenumber", streetnumber->value, 0);
+         freeItem(streetnumber);
+         *housenumber = popItem(tags);
+      }
+      if (!streetnumber)
+      {
+         addItem(tags, "addr:housenumber", conscriptionnumber->value, 10);
+         freeItem(conscriptionnumber);
+         *housenumber = popItem(tags);
+      }
+      if (conscriptionnumber && streetnumber)
+      {
+         char * completenumber = strdup(conscriptionnumber->value);
+         size_t completenumberlength = strlen(completenumber);
+         completenumber = realloc(completenumber, completenumberlength + 2 + strlen(streetnumber->value));
+         *(completenumber + completenumberlength) = '/';
+         strcpy(completenumber + completenumberlength + 1, streetnumber->value);
+         freeItem(conscriptionnumber);
+         freeItem(streetnumber);
+         addItem(tags, "addr:housenumber", completenumber, 0);
+         *housenumber = popItem(tags);
+         free(completenumber);
+      }
+    }
+
+   if (place)
+   {
+      if (placeadmin)
+      {
+         pushItem(extratags, place);
+      } 
+      else
+      {
+         pushItem(places, place);
+      }
+   }
+
    if (placehouse && !listHasData(places))
    {
       addItem(places, "place", "house", 1);
    }
 
-   // Fallback place types - only used if we didn't create something more specific already
-   if (placebuilding && !listHasData(places))
+   /* Fallback place types - only used if we didn't create something more specific already */
+   if (placebuilding && !listHasData(places) && (listHasData(names) || *housenumber || *postcode))
    {
       addItem(places, "building", "yes", 1);
    }
 
-   if (*landuse && !listHasData(places))
+   if (landuse)
    {
-      addItem(places, "landuse", landuse, 1);
+      if (!listHasData(places))
+      {
+          pushItem(places, landuse);
+      }
+      else
+      {
+          freeItem(item);
+      }
    }
 
    if (*postcode && !listHasData(places))
@@ -550,7 +627,7 @@ static int split_tags(struct keyval *tags, unsigned int flags, struct keyval *na
       addItem(places, "place", "postcode", 1);
    }
 
-   // Try to convert everything to an area
+   /* Try to convert everything to an area */
    return 1;
 }
 
@@ -569,7 +646,7 @@ void escape_array_record(char *out, int len, const char *in)
             case '\r': 
             case '\t': 
             case '"': 
-		// This is a bit naughty - we know that nominatim ignored these characters so just drop them now for simplicity
+                /* This is a bit naughty - we know that nominatim ignored these characters so just drop them now for simplicity */
 		*out++ = ' '; count++; break;
             default:   *out++ = *in; count++; break;
         }
@@ -581,135 +658,57 @@ void escape_array_record(char *out, int len, const char *in)
         fprintf(stderr, "%s truncated at %d chars: %s\n%s\n", __FUNCTION__, count, old_in, old_out);
 }
 
+static void delete_unused_classes(char osm_type, osmid_t osm_id, struct keyval *places) {
+    int i,sz, slen;
+    PGresult   *res;
+    char tmp[16];
+    char tmp2[2];
+    char *cls, *clslist = 0;
+    char const *paramValues[2];
+    
+    tmp2[0] = osm_type; tmp2[1] = '\0';
+    paramValues[0] = tmp2;
+    snprintf(tmp, sizeof(tmp), "%" PRIdOSMID, osm_id);
+    paramValues[1] = tmp;
+    res = pgsql_execPrepared(ConnectionDelete, "get_classes", 2, paramValues, PGRES_TUPLES_OK);
 
-static void add_place_v1(char osm_type, osmid_t osm_id, const char *class, const char *type, struct keyval *names, struct keyval *extratags,
-   int adminlevel, const char *housenumber, const char *street, const char *isin, const char *postcode, const char *countrycode, const char *wkt)
-{
-   int first;
-   struct keyval *name;
-   char sql[2048];
+    sz = PQntuples(res);
+    if (sz > 0 && !places) {
+        PQclear(res);
+        /* uncondtional delete of all places */
+        stop_copy();
+        pgsql_exec(Connection, PGRES_COMMAND_OK, "DELETE FROM place WHERE osm_type = '%c' AND osm_id  = %" PRIdOSMID, osm_type, osm_id);
+    } else {
+        for (i = 0; i < sz; i++) {
+            cls = PQgetvalue(res, i, 0);
+            if (!getItem(places, cls)) {
+                if (!clslist) {
+                    clslist = malloc(strlen(cls)+3);
+                    sprintf(clslist, "'%s'", cls);
+                } else {
+                    slen = strlen(clslist);
+                    clslist = realloc(clslist, slen + 4 + strlen(cls));
+                    sprintf(&(clslist[slen]), ",'%s'", cls); 
+                }
+            }
+        }
 
-   /* Output a copy line for this place */
-   sprintf(sql, "\\N\t%c\t%" PRIdOSMID "\t", osm_type, osm_id);
-   copy_data(sql);
+        PQclear(res);
 
-   escape(sql, sizeof(sql), class);
-   copy_data(sql);
-   copy_data("\t");
+        if (clslist) {
+           /* Stop any active copy */
+           stop_copy();
 
-   escape(sql, sizeof(sql), type);
-   copy_data(sql);
-   copy_data("\t");
-
-   /* start name array */
-   if (listHasData(names))
-   {
-      copy_data("{");
-      first = 1;
-      for (name = firstItem(names); name; name = nextItem(names, name))
-      {
-         if (first) first = 0;
-         else copy_data(",");
-
-         copy_data("\"(\\\\\"");
-
-         escape_array_record(sql, sizeof(sql), name->key);
-         copy_data(sql);
-
-         copy_data("\\\\\",\\\\\"");
-
-         escape_array_record(sql, sizeof(sql), name->value);
-         copy_data(sql);
-
-         copy_data("\\\\\")\"");
-      }
-      copy_data("}\t");
-   }
-   else
-   {
-      copy_data("\\N\t");
-   }
-
-   sprintf(sql, "%d\t", adminlevel);
-   copy_data(sql);
-
-   if (housenumber)
-   {
-      escape(sql, sizeof(sql), housenumber);
-      copy_data(sql);
-      copy_data("\t");
-   }
-   else
-   {
-      copy_data("\\N\t");
-   }
-
-   if (street)
-   {
-      escape(sql, sizeof(sql), street);
-      copy_data(sql);
-      copy_data("\t");
-   }
-   else
-   {
-      copy_data("\\N\t");
-   }
-
-   if (isin)
-   {
-      // Skip the leading ',' from the contactination
-      escape(sql, sizeof(sql), isin+1);
-      copy_data(sql);
-      copy_data("\t");
-   }
-   else
-   {
-      copy_data("\\N\t");
-   }
-
-   if (postcode)
-   {
-      escape(sql, sizeof(sql), postcode);
-      copy_data(sql);
-      copy_data("\t");
-   }
-   else
-   {
-      copy_data("\\N\t");
-   }
-
-   if (countrycode)
-   {
-      escape(sql, sizeof(sql), countrycode);
-      copy_data(sql);
-      copy_data("\t");
-   }
-   else
-   {
-     copy_data("\\N\t");
-   }
-
-   copy_data("\\N\t");
-
-   copy_data("\\N\t");
-
-   copy_data("\\N\t");
-
-   copy_data("\\N\t");
-
-   sprintf(sql, "SRID=%d;", SRID);
-   copy_data(sql);
-   copy_data(wkt);
-
-   copy_data("\n");
-
-//fprintf(stderr, "%c %" PRIdOSMID " %s\n", osm_type, osm_id, wkt);
-
-   return;
+           /* Delete all places for this object */
+           pgsql_exec(Connection, PGRES_COMMAND_OK, "DELETE FROM place WHERE osm_type = '%c' AND osm_id = %"
+        PRIdOSMID " and class = any(ARRAY[%s])", osm_type, osm_id, clslist);
+           free(clslist);
+        }
+    }
 }
 
-static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const char *type, struct keyval *names, struct keyval *extratags,
-   int adminlevel, const char *housenumber, const char *street, const char *isin, const char *postcode, const char *countrycode, const char *wkt)
+static void add_place(char osm_type, osmid_t osm_id, const char *class, const char *type, struct keyval *names, struct keyval *extratags,
+   int adminlevel, struct keyval *housenumber, struct keyval *street, const char *isin, struct keyval *postcode, struct keyval *countrycode, const char *wkt)
 {
    int first;
    struct keyval *name;
@@ -760,7 +759,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    if (housenumber)
    {
-      escape(sql, sizeof(sql), housenumber);
+      escape(sql, sizeof(sql), housenumber->value);
       copy_data(sql);
       copy_data("\t");
    }
@@ -771,7 +770,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    if (street)
    {
-      escape(sql, sizeof(sql), street);
+      escape(sql, sizeof(sql), street->value);
       copy_data(sql);
       copy_data("\t");
    }
@@ -782,7 +781,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    if (isin)
    {
-      // Skip the leading ',' from the contactination
+       /* Skip the leading ',' from the contactination */
       escape(sql, sizeof(sql), isin+1);
       copy_data(sql);
       copy_data("\t");
@@ -794,7 +793,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    if (postcode)
    {
-      escape(sql, sizeof(sql), postcode);
+      escape(sql, sizeof(sql), postcode->value);
       copy_data(sql);
       copy_data("\t");
    }
@@ -805,7 +804,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    if (countrycode)
    {
-      escape(sql, sizeof(sql), countrycode);
+      escape(sql, sizeof(sql), countrycode->value);
       copy_data(sql);
       copy_data("\t");
    }
@@ -815,7 +814,7 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
    }
 
    /* extra tags array */
-   if (listHasData(names))
+   if (listHasData(extratags))
    {
       first = 1;
       for (name = firstItem(extratags); name; name = nextItem(extratags, name))
@@ -848,7 +847,6 @@ static void add_place_v2(char osm_type, osmid_t osm_id, const char *class, const
 
    copy_data("\n");
 
-//fprintf(stderr, "%c %" PRIdOSMID " %s\n", osm_type, osm_id, wkt);
 
    return;
 }
@@ -919,20 +917,10 @@ static void add_polygon_error(char osm_type, osmid_t osm_id, const char *class,
 
    copy_error_data("\n");
 
-//fprintf(stderr, "%c %" PRIdOSMID " %s\n", osm_type, osm_id, wkt);
 
    return;
 }
 
-static void add_place(char osm_type, osmid_t osm_id, const char *class, const char *type, struct keyval *names, struct keyval *extratags,
-   int adminlevel, const char *housenumber, const char *street, const char *isin, const char *postcode, const char *countrycode, const char *wkt)
-{
-   if (Options->enable_hstore)
-      add_place_v2(osm_type, osm_id, class, type, names, extratags, adminlevel, housenumber, street, isin, postcode, countrycode, wkt);
-   else
-      add_place_v1(osm_type, osm_id, class, type, names, extratags, adminlevel, housenumber, street, isin, postcode, countrycode, wkt);
-   return;
-}
 
 static void delete_place(char osm_type, osmid_t osm_id)
 {
@@ -952,7 +940,6 @@ static int gazetteer_out_start(const struct output_options *options)
 
    /* Connection to the database */
    Connection = PQconnectdb(options->conninfo);
-   ConnectionError = PQconnectdb(options->conninfo);
 
    /* Check to see that the backend connection was successfully made */
    if (PQstatus(Connection) != CONNECTION_OK)
@@ -976,21 +963,18 @@ static int gazetteer_out_start(const struct output_options *options)
       pgsql_exec(Connection, PGRES_COMMAND_OK, "DROP FUNCTION IF EXISTS get_connected_ways(integer[])");
 
       /* Create types and functions */
-      if (!Options->enable_hstore)
-      {
-            pgsql_exec(Connection, PGRES_COMMAND_OK, CREATE_KEYVALUETYPE_TYPE, "", "");
-      }
-      pgsql_exec(Connection, PGRES_COMMAND_OK, CREATE_WORDSCORE_TYPE, Options->tblsmain_data);
+      pgsql_exec(Connection, PGRES_COMMAND_OK, CREATE_KEYVALUETYPE_TYPE);
+      pgsql_exec(Connection, PGRES_COMMAND_OK, CREATE_WORDSCORE_TYPE);
 
       /* Create the new table */
       if (Options->tblsmain_data)
       {
-          pgsql_exec(Connection, PGRES_COMMAND_OK, Options->enable_hstore ? 
-            V2_CREATE_PLACE_TABLE : CREATE_PLACE_TABLE, "TABLESPACE", Options->tblsmain_data);
+          pgsql_exec(Connection, PGRES_COMMAND_OK,
+                      CREATE_PLACE_TABLE, "TABLESPACE", Options->tblsmain_data);
       }
       else
       {
-          pgsql_exec(Connection, PGRES_COMMAND_OK, Options->enable_hstore ? V2_CREATE_PLACE_TABLE : CREATE_PLACE_TABLE, "", "");
+          pgsql_exec(Connection, PGRES_COMMAND_OK, CREATE_PLACE_TABLE, "", "");
       }
       if (Options->tblsmain_index)
       {
@@ -1003,6 +987,15 @@ static int gazetteer_out_start(const struct output_options *options)
 
       pgsql_exec(Connection, PGRES_TUPLES_OK, "SELECT AddGeometryColumn('place', 'geometry', %d, 'GEOMETRY', 2)", SRID);
       pgsql_exec(Connection, PGRES_COMMAND_OK, "ALTER TABLE place ALTER COLUMN geometry SET NOT NULL");
+   } else {
+      ConnectionDelete = PQconnectdb(options->conninfo);
+      if (PQstatus(ConnectionDelete) != CONNECTION_OK)
+      { 
+          fprintf(stderr, "Connection to database failed: %s\n", PQerrorMessage(ConnectionDelete));
+          exit_nicely();
+      }
+
+      pgsql_exec(ConnectionDelete, PGRES_COMMAND_OK, "PREPARE get_classes (CHAR(1), " POSTGRES_OSMID_TYPE ") AS SELECT class FROM place WHERE osm_type = $1 and osm_id = $2");
    }
 
    /* Setup middle layer */
@@ -1016,8 +1009,6 @@ static int gazetteer_out_start(const struct output_options *options)
 static void gazetteer_out_stop(void)
 {
    /* Process any remaining ways and relations */
-//   Options->mid->iterate_ways( gazetteer_out_way );
-//   Options->mid->iterate_relations( gazetteer_process_relation );
 
    /* No longer need to access middle layer */
    Options->mid->commit();
@@ -1025,14 +1016,17 @@ static void gazetteer_out_stop(void)
 
    /* Stop any active copy */
    stop_copy();
-   //stop_error_copy();
    if (hLog) fclose(hLog);
 
    /* Commit transaction */
    pgsql_exec(Connection, PGRES_COMMAND_OK, "COMMIT");
 
-   /* Analyse the table */
-   //pgsql_exec(Connection, PGRES_COMMAND_OK, "ANALYZE place");
+
+   PQfinish(Connection);
+   if (ConnectionDelete)
+       PQfinish(ConnectionDelete);
+   if (ConnectionError)
+       PQfinish(ConnectionError);
 
    return;
 }
@@ -1042,21 +1036,20 @@ static void gazetteer_out_cleanup(void)
    return;
 }
 
-static int gazetteer_add_node(osmid_t id, double lat, double lon, struct keyval *tags)
+static int gazetteer_process_node(osmid_t id, double lat, double lon, struct keyval *tags, int delete_old)
 {
    struct keyval names;
    struct keyval places;
    struct keyval extratags;
    struct keyval *place;
    int adminlevel;
-   char * housenumber;
-   char * street;
+   struct keyval * housenumber;
+   struct keyval * street;
    char * isin;
-   char * postcode;
-   char * countrycode;
+   struct keyval * postcode;
+   struct keyval * countrycode;
    char wkt[128];
 
-//fprintf(stderr, "node\n");
 
    /* Split the tags */
    split_tags(tags, TAGINFO_NODE, &names, &places, &extratags, &adminlevel, &housenumber, &street, &isin, &postcode, &countrycode);
@@ -1064,8 +1057,11 @@ static int gazetteer_add_node(osmid_t id, double lat, double lon, struct keyval
    /* Feed this node to the middle layer */
    Options->mid->nodes_set(id, lat, lon, tags);
 
+   if (delete_old)
+       delete_unused_classes('N', id, &places);
+
    /* Are we interested in this item? */
-   if (listHasData(&names) || listHasData(&places))
+   if (listHasData(&places))
    {
       sprintf(wkt, "POINT(%.15g %.15g)", lon, lat);
       for (place = firstItem(&places); place; place = nextItem(&places, place))
@@ -1074,30 +1070,39 @@ static int gazetteer_add_node(osmid_t id, double lat, double lon, struct keyval
       }
    }
 
+   if (housenumber) freeItem(housenumber);
+   if (street) freeItem(street);
    if (isin) free(isin);
+   if (postcode) freeItem(postcode);
+   if (countrycode) freeItem(countrycode);
 
    /* Free tag lists */
    resetList(&names);
    resetList(&places);
+   resetList(&extratags);
 
    return 0;
 }
 
-static int gazetteer_add_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *tags)
+static int gazetteer_add_node(osmid_t id, double lat, double lon, struct keyval *tags)
+{
+    return gazetteer_process_node(id, lat, lon, tags, 0);
+}
+
+static int gazetteer_process_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *tags, int delete_old)
 {
    struct keyval names;
    struct keyval places;
    struct keyval extratags;
    struct keyval *place;
    int adminlevel;
-   char * housenumber;
-   char * street;
+   struct keyval * housenumber;
+   struct keyval * street;
    char * isin;
-   char * postcode;
-   char * countrycode;
+   struct keyval * postcode;
+   struct keyval * countrycode;
    int area;
 
-//fprintf(stderr, "way\n");
 
    /* Split the tags */
    area = split_tags(tags, TAGINFO_WAY, &names, &places, &extratags, &adminlevel, &housenumber, &street, &isin, &postcode, &countrycode);
@@ -1105,8 +1110,11 @@ static int gazetteer_add_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *t
    /* Feed this way to the middle layer */
    Options->mid->ways_set(id, ndv, ndc, tags, 0);
 
+   if (delete_old)
+       delete_unused_classes('W', id, &places);
+
    /* Are we interested in this item? */
-   if (listHasData(&names) || listHasData(&places))
+   if (listHasData(&places))
    {
       struct osmNode *nodev;
       int nodec;
@@ -1132,56 +1140,75 @@ static int gazetteer_add_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *t
       free(nodev);
    }
 
+   if (housenumber) freeItem(housenumber);
+   if (street) freeItem(street);
    if (isin) free(isin);
+   if (postcode) freeItem(postcode);
+   if (countrycode) freeItem(countrycode);
 
    /* Free tag lists */
    resetList(&names);
    resetList(&places);
+   resetList(&extratags);
 
    return 0;
 }
 
-static int gazetteer_add_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags)
+static int gazetteer_add_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *tags)
+{
+    return gazetteer_process_way(id, ndv, ndc, tags, 0);
+}
+
+static int gazetteer_process_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags, int delete_old)
 {
    struct keyval names;
    struct keyval places;
    struct keyval extratags;
+   struct keyval *place;
    int adminlevel;
-   char * housenumber;
-   char * street;
+   struct keyval * housenumber;
+   struct keyval * street;
    char * isin;
-   char * postcode;
-   char * countrycode;
-   int area, wkt_size;
+   struct keyval * postcode;
+   struct keyval * countrycode;
+   int wkt_size;
    const char *type;
-   const char *boundary;
 
    type = getItem(tags, "type");
-   if (!type)
+   if (!type) {
+      if (delete_old) delete_unused_classes('R', id, 0); 
       return 0;
+   }
 
    if (!strcmp(type, "associatedStreet") || !strcmp(type, "relatedStreet"))
    {
       Options->mid->relations_set(id, members, member_count, tags);
+      if (delete_old) delete_unused_classes('R', id, 0); 
       return 0;
    }
 
-   if (strcmp(type, "boundary") && strcmp(type, "multipolygon"))
+   if (strcmp(type, "boundary") && strcmp(type, "multipolygon")) {
+      if (delete_old) delete_unused_classes('R', id, 0); 
       return 0;
+   }
 
-   boundary = getItem(tags, "boundary");
-   if (!boundary) boundary = "administrative";
+   Options->mid->relations_set(id, members, member_count, tags);
 
    /* Split the tags */
-   area = split_tags(tags, TAGINFO_AREA, &names, &places, &extratags, &adminlevel, &housenumber, &street, &isin, &postcode, &countrycode);
+   split_tags(tags, TAGINFO_AREA, &names, &places, &extratags, &adminlevel, &housenumber, &street, &isin, &postcode, &countrycode);
 
-   if (listHasData(&names))
+   if (delete_old)
+       delete_unused_classes('R', id, &places);
+
+   if (listHasData(&places))
    {
       /* get the boundary path (ways) */
       int i, count;
       int *xcount = malloc( (member_count+1) * sizeof(int) );
       struct keyval *xtags  = malloc( (member_count+1) * sizeof(struct keyval) );
       struct osmNode **xnodes = malloc( (member_count+1) * sizeof(struct osmNode*) );
+      osmid_t *xid;
+      osmid_t *xid2 = malloc( (member_count+1) * sizeof(osmid_t) );
 
       count = 0;
       for (i=0; i<member_count; i++)
@@ -1189,12 +1216,12 @@ static int gazetteer_add_relation(osmid_t id, struct member *members, int member
          /* only interested in ways */
          if (members[i].type != OSMTYPE_WAY)
             continue;
-
-         initList(&(xtags[count]));
-         if (Options->mid->ways_get( members[i].id, &(xtags[count]), &(xnodes[count]), &(xcount[count])))
-            continue;
+         xid2[count] = members[i].id;
          count++;
       }
+
+      count = Options->mid->ways_get_list(xid2, count, &xid, xtags, xnodes, xcount);
+
       xnodes[count] = NULL;
       xcount[count] = 0;
 
@@ -1204,14 +1231,14 @@ static int gazetteer_add_relation(osmid_t id, struct member *members, int member
          char *wkt = get_wkt(i);
          if (strlen(wkt) && (!strncmp(wkt, "POLYGON", strlen("POLYGON")) || !strncmp(wkt, "MULTIPOLYGON", strlen("MULTIPOLYGON"))))
          {
-            if (Options->enable_hstore)
-                add_place('R', id, "boundary", boundary, &names, &extratags, adminlevel, housenumber, street, isin, postcode, countrycode, wkt);
-            else
-                add_place('R', id, "boundary", "adminitrative", &names, &extratags, adminlevel, housenumber, street, isin, postcode, countrycode, wkt);
+             for (place = firstItem(&places); place; place = nextItem(&places, place))
+             {
+                add_place('R', id, place->key, place->value, &names, &extratags, adminlevel, housenumber, street, isin, postcode, countrycode, wkt);
+             }
          }
          else
          {
-            //add_polygon_error('R', id, "boundary", "adminitrative", &names, countrycode, wkt);
+             /* add_polygon_error('R', id, "boundary", "adminitrative", &names, countrycode, wkt); */
          }
          free(wkt);
       }
@@ -1223,20 +1250,32 @@ static int gazetteer_add_relation(osmid_t id, struct member *members, int member
          free( xnodes[i] );
       }
 
+      free(xid);
+      free(xid2);
       free(xcount);
       free(xtags);
       free(xnodes);
    }
 
+   if (housenumber) freeItem(housenumber);
+   if (street) freeItem(street);
    if (isin) free(isin);
+   if (postcode) freeItem(postcode);
+   if (countrycode) freeItem(countrycode);
 
    /* Free tag lists */
    resetList(&names);
    resetList(&places);
+   resetList(&extratags);
 
    return 0;
 }
 
+static int gazetteer_add_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags) 
+{
+    return gazetteer_process_relation(id, members, member_count, tags, 0);
+}
+
 static int gazetteer_delete_node(osmid_t id)
 {
    /* Make sure we are in slim mode */
@@ -1283,21 +1322,21 @@ static int gazetteer_modify_node(osmid_t id, double lat, double lon, struct keyv
 {
    require_slim_mode();
    Options->mid->nodes_delete(id);
-   return gazetteer_add_node(id, lat, lon, tags);
+   return gazetteer_process_node(id, lat, lon, tags, 1);
 }
 
 static int gazetteer_modify_way(osmid_t id, osmid_t *ndv, int ndc, struct keyval *tags)
 {
    require_slim_mode();
    Options->mid->ways_delete(id);
-   return gazetteer_add_way(id, ndv, ndc, tags);
+   return gazetteer_process_way(id, ndv, ndc, tags, 1);
 }
 
 static int gazetteer_modify_relation(osmid_t id, struct member *members, int member_count, struct keyval *tags)
 {
    require_slim_mode();
    Options->mid->relations_delete(id);
-   return gazetteer_add_relation(id, members, member_count, tags);
+   return gazetteer_process_relation(id, members, member_count, tags, 1);
 }
 
 struct output_t out_gazetteer = {
