@@ -21,39 +21,39 @@ static int null_out_start(const struct output_options *opt UNUSED) {
 static void null_out_stop() {
 }
 
-static int null_add_node(int a UNUSED, double b UNUSED, double c UNUSED, struct keyval *k UNUSED) {
+static int null_add_node(osmid_t a UNUSED, double b UNUSED, double c UNUSED, struct keyval *k UNUSED) {
   return 0;
 }
 
-static int null_add_way(int a UNUSED, int *b UNUSED, int c UNUSED, struct keyval *k UNUSED) {
+static int null_add_way(osmid_t a UNUSED, osmid_t *b UNUSED, int c UNUSED, struct keyval *k UNUSED) {
   return 0;
 }
 
-static int null_add_relation(int a UNUSED, struct member *b UNUSED, int c UNUSED, struct keyval *k UNUSED) {
+static int null_add_relation(osmid_t a UNUSED, struct member *b UNUSED, int c UNUSED, struct keyval *k UNUSED) {
   return 0;
 }
 
-static int null_delete_node(int i UNUSED) {
+static int null_delete_node(osmid_t i UNUSED) {
   return 0;
 }
 
-static int null_delete_way(int i UNUSED) {
+static int null_delete_way(osmid_t i UNUSED) {
   return 0;
 }
 
-static int null_delete_relation(int i UNUSED) {
+static int null_delete_relation(osmid_t i UNUSED) {
   return 0;
 }
 
-static int null_modify_node(int a UNUSED, double b UNUSED, double c UNUSED, struct keyval * k UNUSED) {
+static int null_modify_node(osmid_t a UNUSED, double b UNUSED, double c UNUSED, struct keyval * k UNUSED) {
   return 0;
 }
 
-static int null_modify_way(int a UNUSED, int * b UNUSED, int c UNUSED, struct keyval * k UNUSED) {
+static int null_modify_way(osmid_t a UNUSED, osmid_t * b UNUSED, int c UNUSED, struct keyval * k UNUSED) {
   return 0;
 }
 
-static int null_modify_relation(int a UNUSED, struct member * b UNUSED, int c UNUSED, struct keyval * k UNUSED) {
+static int null_modify_relation(osmid_t a UNUSED, struct member * b UNUSED, int c UNUSED, struct keyval * k UNUSED) {
   return 0;
 }
 
