@@ -40,6 +40,7 @@ struct output_options {
   int expire_tiles_zoom_min;	/* Minimum zoom level for tile expiry list */
   const char *expire_tiles_filename;	/* File name to output expired tiles list to */
   int enable_hstore; /* add an additional hstore column with objects key/value pairs */
+  int enable_hstore_index; /* add an index on the hstore column */
   int enable_multi; /* Output multi-geometries intead of several simple geometries */
   const char** hstore_columns; /* list of columns that should be written into their own hstore column */
   int n_hstore_columns; /* number of hstore columns */
@@ -49,6 +50,10 @@ struct output_options {
   int num_procs;
   int droptemp; /* drop slim mode temp tables after act */
   int unlogged; /* use unlogged tables where possible */
+  int hstore_match_only; /* only copy rows that match an explicitly listed key */
+  int flat_node_cache_enabled;
+  int excludepoly;
+  const char *flat_node_file;
 };
 
 struct output_t {
@@ -57,10 +62,6 @@ struct output_t {
     void (*stop)();
     void (*cleanup)(void);
     void (*close)(int stopTransaction);
-//    void (*process)(struct middle_t *mid);
-//    int (*node)(osmid_t id, struct keyval *tags, double node_lat, double node_lon);
-//    int (*way)(osmid_t id, struct keyval *tags, struct osmNode *nodes, int count);
-//    int (*relation)(osmid_t id, struct keyval *rel_tags, struct osmNode **nodes, struct keyval **tags, int *count);
 
     int (*node_add)(osmid_t id, double lat, double lon, struct keyval *tags);
     int (*way_add)(osmid_t id, osmid_t *nodes, int node_count, struct keyval *tags);
